/*
  Library support for FlexRadio 6000 Rigs
  Copyright (C)2015 Vincenzo Stefanazzi - IW7DMH. All right reserved

  This library is free software; you can redistribute it and/or
  modify it under the terms of the CC BY-NC-SA 3.0 license.
  https://creativecommons.org/licenses/by-nc-sa/3.0/
  
  The license applies to all part of the library including the 
  examples and tools supplied with the library.
*/
#include <FlexRig.h>
FlexRig::FlexRig() {
	C=0;
	sprintf(handle,"3144B418"); //dummy initialization for offline test
	radio=Radio();
	transmit=Transmit();
	interlock=Interlock();
	panadapter[0]=Panadapter();
	panadapter[1]=Panadapter();
	waterfall[0]=Waterfall();
	waterfall[1]=Waterfall();
	slice[0]=Slice();
	slice[1]=Slice();
}


void FlexRig::ping() {
	send("ping");
}


void FlexRig::send(String cmd) {
	send(cmd,0);
}


void FlexRig::send(String cmd, int parserId) {
    char c[60];
	//C++;
	sprintf(c, "C%d|%s\n",C++,cmd.c_str());
	//String cmd2=String(c);
	//cmd2 += cmd + "\n";

	//Protect critical section from interrupt events
	noInterrupts();
	  //addToResponseList(C,parserId);
	  //sending command
	  client.println(c); 
	interrupts();
	//Serial.print("->"); 
	Serial.print(c); 
	//Serial.println("<-"); //Debug
}


void FlexRig::readEthernetData() {
	
	//reading response
	while (client.available()) {
		//Serial.print("numero di byte:");Serial.println(client.available());
		char c=client.read();
		if (c==0) {
		    //Serial.println("0!!!");
			continue;
		}
		//Serial.print(c); 
		
        if (c=='\n') { 
		    
			buffer.trim();

			//Serial.print("===Buffer address: "); Serial.println((long)&buffer,HEX);

			switch (buffer[0]) {
			case 'R':
			  //Serial.println("==RESPONSE==");
			  parseReplay(buffer);
			  break;
			case 'S':
			  //Serial.println("==STATUS==");
			  parseStatus(buffer);
			  break;
			case 'M':
			  //Serial.println("==MESSAGE==");
			  parseMessage(buffer);
			  break;
			default:
			  Serial.println("\n==UNKNOWN==");
			  Serial.print("Buffer lenght:"); Serial.println(buffer.length());
			  int i=0;
			  for (i=0;i<buffer.length();i++) {
			     Serial.print(buffer[0]); Serial.print("-");Serial.println(buffer[0],DEC);
			  }
			  Serial.print(">"); Serial.print(buffer);Serial.println("<");
			}
			buffer=""; 
		} else {
			buffer+=c;
		} 
		
	} 

}

void FlexRig::parseStatus(String msg){
	//Serial.print("FlexRig::parseStatus(msg): "); 
	//Serial.println(msg);
	//if (msg.indexOf(handle)) {
	    int s=msg.indexOf(' ');
		int h=msg.indexOf('|')+1;
		String type=msg.substring(h,s);
		//Serial.println(type);
		if (type.equals("radio")){
		    radio.updateStatus(msg.substring(s+1));
			return;
		}
		if (type.equals("transmit")){
		    transmit.updateStatus(msg.substring(s+1));
			return;
		}
		if (type.equals("interlock")){
		    interlock.updateStatus(msg.substring(s+1));
			return;
		} 
		if (type.equals("display")){
		    int s1=msg.indexOf(' ',s+1);
			String obj=msg.substring(18,s1);
			//Serial.print("==="); Serial.print(obj); Serial.println("===");
		    //Serial.print("===pan (position) "); Serial.println(s1);
        	int id=(msg.substring(s1+3,s1+11)).toInt();
			if (obj.equals("pan")) {
				//Serial.print("===pan (before) "); Serial.println(id);
        		id=id % 40000000;	
				//Serial.print("===pan (after) "); Serial.println(id);
				panadapter[id].updateStatus(msg.substring(s1+12));
			}
			
			if (obj.equals("waterfall")) {
			    //Serial.print("===waterfall (before) "); Serial.println(id);
        		id=id % 42000000;
				//Serial.print("===waterfall (after) "); Serial.println(id);
			    waterfall[id].updateStatus(msg.substring(s1+12));
			}
			
			return;
		}
		if (type.equals("slice")){
	        //Serial.print("msg=");Serial.println(msg);
    		int s1=msg.indexOf(' ',s);
		    //Serial.print("s1:"); Serial.println(s1);
			int id=(msg.substring(s1+1,s1+2)).toInt();
			//Serial.print("slice id="); Serial.println(id);
			slice[id].updateStatus(msg.substring(s1+1));
			return;
		}
		
		if (type.equals("eq")){
			//Serial.println("eq");
			eq.updateStatus(msg.substring(s+1));
			return;
		}
		
		
		if (type.equals("waveform")){
			Serial.println("waveform");
			return;
		}
		

		/////////////////////////////////////////////////////////////Serial.print("FlexRig::parseStatus() - Unhandled type->"); Serial.println(type + "<");
	//}
}

void FlexRig::parseMessage(String msg){
	Serial.println(msg); 
}


void FlexRig::connect() {

  Serial.print("connecting to Flex rig:");
  Serial.print(ipAddress[0]); Serial.print(".");
  Serial.print(ipAddress[1]); Serial.print(".");
  Serial.print(ipAddress[2]); Serial.print(".");
  Serial.print(ipAddress[3]);  Serial.println(".");

  if (client.connect(ipAddress, UDP_DISCOVERING_PORT)) {
    connected=true;
    Serial.println("Connected");
	//ping();
	//String buffer;

	delay(1000);
	while (client.available()) {
		char c=client.read();
	    if (c=='\n') {
			switch (buffer[0]) {
			case 'V':
			  Serial.println("==VERSION (connect)==");
			  for (int i=1;i<buffer.length();i++) {
				softVersion[i-1]=buffer[i];
				Serial.print(softVersion[i-1]);
			  }
			  Serial.println();
			  break;
			case 'H':
			  Serial.println("==HANDLE (connect)==");
			  for (int i=1;i<buffer.length();i++) {
				handle[i-1]=buffer[i];
				Serial.print(handle[i-1]);
			  }
			  Serial.println();
			  break;
            /*case 'S':
			  Serial.println("==STATUS (connect)==");
			  parseStatus(buffer);
			  break;
			case 'M':
			  Serial.println("==MESSAGE(connect)==");
			  parseMessage(buffer);
			  break;
			*/
		}
			buffer="";
		} else {
			buffer+=c;
		}
	}
	configureClient();

	//
	//startReceiver();
	//
  } else {
    Serial.println("failed");
  }

}

void FlexRig::initBuffer(){
    //dummy initializazion with 1024 bytes
	/*
	buffer="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	buffer+="                                                                ";
	*/
}

boolean FlexRig::isAFlex(char* model, char* serial, char* serialKey) {
  boolean found=true;
  char pattern[7] = "FLEX-6";

  for (int i=0;i<6;i++) {
    if (pattern[i]!=model[i]) {
		//Serial.print(model[i]); Serial.print("-"); Serial.println(pattern[i]);
      found=false;
      break;
    }
  }
  
  if (found && serialKey!=NULL) {  //looking for a specific rig
	for (int i=0;i<19;i++) {
		if (serial[i]!=serialKey[i]) {
			//Serial.print(serial[i]); Serial.print("-"); Serial.println(serialKey[i]);
		  found=false;
		  break;
		}
    }  
  }
  //Serial.print("FlexRig::isAFlex():"); Serial.println(found);

  return found;
}

FlexRig FlexRig::findAFlex(char* serialKey) {

  EthernetUDP Udp;
  char packetBuffer[UDP_TX_PACKET_MAX_SIZE]; //buffer to hold incoming packet,
  Udp.begin(UDP_DISCOVERING_PORT);
  FlexRig f=FlexRig();
  f.initBuffer();

  Serial.println("===Looking for Flex Rig (UDP)===");

  boolean found=false;
  while (!found) {

      int packetSize = Udp.parsePacket();

      //Serial.print("Received packet - size ");
      //Serial.println(packetSize);


      //if ((packetSize==UDP_TX_PACKET_MAX_SIZE) || (packetSize==228))
      if (packetSize>=228)
      {
        // read the packet into packetBufffer
        Udp.read(packetBuffer,packetSize);

        char model[10];
        for (int i =0; i < 10; i++) {
          model[i]=packetBuffer[69+i];
          //Serial.println(model[i]);
	    }
		
		char serialTmpVal[20];
		Serial.print("=====>"); 
        for (int i=0; i < 20; i++) {
            serialTmpVal[i]=packetBuffer[86+i];
            Serial.print(packetBuffer[86+i]);
        }
        Serial.println();
		
        if (isAFlex(model,serialTmpVal,serialKey)) {

          Serial.println("===========Flex Rig=============");

          Serial.print("IP Address:");
          IPAddress remote = Udp.remoteIP();
          for (int i =0; i < 4; i++) {
            Serial.print(remote[i], DEC);
            f.ipAddress[i]=remote[i];       //Storing data into Flex Struct
            if (i < 3) Serial.print(".");
          }
          Serial.println();


          Serial.print("Model Name:");
          for (int i=69; i < 78; i++) {
            f.modelName[i-69]=packetBuffer[i]; //Storing data into Flex Struct
            Serial.print(packetBuffer[i]);
          }
          Serial.println();


          Serial.print("Serial:");
		  for (int i=86; i < 105; i++) {
            f.serial[i-86]=packetBuffer[i]; //Storing data into Flex Struct
            Serial.print(packetBuffer[i]);
          }
          Serial.println();

          Serial.print("Version:");
 		  for (int i=114; i < 122; i++) {
		    f.softVersion[i-114]=packetBuffer[i]; //Storing data into Flex Struct
		    Serial.print(packetBuffer[i]);
		  }
          Serial.println();

          Serial.print("NickName:");
          for (int i=132; i < 145; i++) {
            f.nickName[i-132]=packetBuffer[i]; //Storing data into Flex Struct
            Serial.print(packetBuffer[i]);
          }
          Serial.println();

          found=true;
        }

      }

      delay(100);

  } //main loop simulation - it is executed only once until a Flex rig is not found

   return f;
}

void FlexRig::configureClient() {
	delay(500);
	//send("client program DMH_CONTROL");
    send("sub tx all");
	//S69A3CDEB|transmit freq=14.03619 lo=100 hi=2900 rfpower=3 tunepower=10 am_carrier_level=100 vox_enable=0 vox_level=50 vox_delay=250 mic_level=40 mic_selection=MIC mic_boost=1 mic_bias=0 mic_acc=0 compander=1 compander_level=70 dax=0 pitch=540 speed=23 iambic=1 iambic_mode=1 cwl_enabled=0 swap_paddles=0 break_in=1 break_in_delay=9 sidetone=1 monitor=0 mon_gain_cw=76 mon_gain_sb=80 mon_pan_cw=50 mon_pan_sb=50 tune=0 met_in_rx=0 hwalc_enabled=0 inhibit=0 speech_processor_enable=1 speech_processor_level=0 show_tx_in_waterfall=0
	send("sub atu all");
	//S69A3CDEB|atu status=NONE atu atu_enabled=0
	//send("sub meter all");
	//elenco dei meter disponibili
	send("sub pan all");
	//S69A3CDEB|display pan 0x40000000 x_pixels=1170 y_pixels=449 center=14.03499 bandwidth=0.00984 min_dbm=-143.75 max_dbm=-58.75 fps=23 average=10 weighted_average=0 rfgain=0 rxant=ANT1 wide=0 loopa=0 loopb=0 band=20 daxiq=1 daxiq_rate=48000 capacity=16 available=14 waterfall=42000000 min_bw=0.004919999957085 max_bw=7.372800292969 xvtr= pre= ant_list=ANT1,ANT2,XVTR
    //S69A3CDEB|display waterfall 0x42000000 x_pixels=1170 center=14.03499 bandwidth=0.00984 line_duration=100 rfgain=0 rxant=ANT1 wide=0 loopa=0 loopb=0 band=20 daxiq=0 daxiq_rate=0 capacity=16 available=14 panadapter=40000000 color_gain=27 auto_black=1 black_level=0 gradient_index=0 xvtr=
	send("sub slice all");
	//S69A3CDEB|slice 0 in_use=1 RF_frequency=14.036190 rit_on=0 rit_freq=0 xit_on=0 xit_freq=0 rxant=ANT1 mode=CW wide=0 filter_lo=-200 filter_hi=200 step=10 step_list=1,5,10,50,100,200,400 agc_mode=med agc_threshold=65 agc_off_level=10 pan=0x40000000 txant=ANT1 loopa=0 loopb=0 qsk=1 dax=1 dax_clients=0 lock=0 tx=1 dax_tx=0 active=1 audio_gain=50 audio_pan=50 audio_mute=0 record=0 play=disabled record_time=0.0 anf=0 anf_level=0 nr=0 nr_level=50 nb=0 nb_level=50 apf=0 apf_level=0 diversity=0 diversity_parent=0 diversity_child=0 diversity_index=1342177293 ant_list=ANT1,ANT2,XVTR
	//send("sub gps all");
	//send("sub audio_stream all");
	//send("sub cwx all");
	//send("sub xvtr all");
	//send("info");
	//send("version");
	send("ant list",1);
	send("mic list",2);
	//send("profile global info");
	//send("profile tx info");
	//send("profile mic info");
	//send("profile display info");
	//send("client udpport 4991");
    send("eq RXsc info");
	send("eq TXsc info");
	delay(500);
}

void FlexRig::startReceiver() {
    send("display pan c 7.020000 1 800 400");
	send("slice C 7.020000 ant1 cw");
}

void FlexRig::stopReceiver() {
    send("slice r 0");
	send("slice r 1");
	send("display pan r 0x40000000");
	send("display pan r 1x40000000");
}

void FlexRig::addToResponseList(int id, int parserId) {

  if((front==0&&rear==RESPONSE_PACKET_LIST_SIZE-1)||(front>0&&rear==front-1)) {
    Serial.println("\nResponse List overflow");
    return;
  } else {
    if(rear==RESPONSE_PACKET_LIST_SIZE-1&&front>0) {
      rear=0;
    } else {
      if((front==0&&rear==-1)||(rear!=front-1))
      rear++;
    }
  }
  packetElement[rear].id=id;
  packetElement[rear].parserId=parserId;
  size++;

}

rPacket FlexRig::removeFromResponseList()
{
  rPacket element;
  if((front==0)&&(rear==-1)) {
      Serial.println("Response List underflow\n");
      return element;
  }
  if(front==rear){
      element=packetElement[front];
      rear=-1;
      front=0;
  } else {
      if(front==RESPONSE_PACKET_LIST_SIZE-1) {
           element=packetElement[front];
           front=0;
      } else {
        element=packetElement[front];
        front++;
      }
  }
  size--;
  return element;
}

void FlexRig::parseReplay(String msg) {

  Serial.println(msg);
  int id = msg.substring(1, msg.indexOf("|")).toInt();

  while (size>0) {
     rPacket tmp = removeFromResponseList();
     if (tmp.id==id) {
        switch (tmp.parserId) {
			case 0:
			  //parseDummy(msg.substring(msg.lastIndexOf("|")+1));
			  break;
			case 1:
			  parseAntennaList(msg.substring(msg.lastIndexOf("|")+1));
			  break;
			case 2:
			  parseMicList(msg.substring(msg.lastIndexOf("|")+1));
			  break;

			default:
			 NULL; ///////////////////////////////////////Serial.println(msg);
	   }
       break;
     }
  }
}

void FlexRig::parseDummy(String msg) {
  //Serial.println("=== Dummy ===");
  //Serial.println(msg+"<");
  //Serial.print("===front:"); Serial.print(front);Serial.print(" - rear:"); Serial.print(rear);Serial.print(" - size:"); Serial.print(size);
  //Serial.println();
}

void FlexRig::parseAntennaList(String msg) {
  //Serial.println("=== Antenna List ===");
  Serial.println(msg);
}

void FlexRig::parseMicList(String msg) {
  //Serial.println("=== Mic List ===");
  Serial.println(msg);
}


void FlexRig::updateObject(int objectId,int value,int activeSlice,int activePanadapter,int activeWaterfall) {
	
	int i=(int)objectId/1000;
	/* do not use this code - it locks the console!!!!
	Serial.print("updateStatus id=");Serial.print(objectId);
	Serial.print(" - value=");Serial.print(value);
	Serial.print(" - activeSlice=");Serial.println(activeId);
	*/
	switch(i) {
		case 1: //Radio
		        this->radio.updateObject(objectId,value);
				break;
		case 2: //Interlock
		        this->interlock.updateObject(objectId,value);
				break;
		case 3: //Transmit
		        this->transmit.updateObject(objectId,value);
				break;
		case 4: //Waterfall (0/1)
				this->waterfall[activeWaterfall].updateObject(objectId,value);
				break;
		case 5: //Panadapter (0/1)
		        this->panadapter[activePanadapter].updateObject(objectId,value);
				break;
		case 6: //Slice (0/1)
		        this->slice[activeSlice].updateObject(objectId,value);
				break;
		case 7: //Eq
		        this->eq.updateObject(objectId,value);
				break;
		default:
			    Serial.print("FlexRig::updateObject - Unhandled Status:");Serial.println(i);
	}
    
}

void  FlexRig::sendAllCommands() {
	while (cSize>0) {
		//Serial.println(removeFromCommandList());
		send(removeFromCommandList());
	}
}

/*
 * Add a command in the list
 */
void FlexRig::addToCommandList(String entry) {

  if((cFront==0&&cRear==COMMAND_LIST_SIZE-1)||(cFront>0&&cRear==cFront-1)) {
    Serial.println("\Command List overflow");
    return;
  } else {
    if(cRear==COMMAND_LIST_SIZE-1&&cFront>0) {
      cRear=0;
    } else {
      if((cFront==0&&cRear==-1)||(cRear!=cFront-1))
      cRear++;
    }
  }
  CommandEntry[cRear]=entry;
  cSize++;
}

/*
 * Return (and remove) the first command in the list
 */
String FlexRig::removeFromCommandList()
{
  String element;
  if((cFront==0)&&(cRear==-1)) {
      Serial.println("\nCommand List underflow");
      return element;
  }
  if(cFront==cRear){
      element=CommandEntry[cFront];
      cRear=-1;
      cFront=0;
  } else {
      if(cFront==COMMAND_LIST_SIZE-1) {
           element=CommandEntry[cFront];
           cFront=0;
      } else {
        element=CommandEntry[cFront];
        cFront++;
      }
  }
  cSize--;
  //Serial.print("->"); Serial.println(element);
  return element;
}

